"
 COPYRIGHT (c) 1999 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"


ListModelView subclass:#SelectionInListModelView
	instanceVariableNames:'selection multipleSelectOk actionBlock doubleClickActionBlock
		clickLine highlightMode useIndex hilightFgColor hilightBgColor
		hilightLevel hilightFrameColor hilightStyle dragAccessPoint
		dropTarget dropSource'
	classVariableNames:'DefaultHilightStyle DefaultHilightBackgroundColor
		DefaultHilightForegroundColor DefaultHilightLevel
		DefaultHilightFrameColor'
	poolDictionaries:''
	category:'Views-Lists'
!

!SelectionInListModelView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1999 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

!

documentation
"
    SelectionInListModelView is mostly like SelectionInListView,
    but derives from the ListModelView and thus the list is kept
    by the model.

    [Instance variables:]
        selection               <misc>     the current selection. nil, a number or collection of numbers
        multipleSelectOk        <Boolean>  allow/disallow multiple selections( default:false )
        actionBlock             <Block>    action evaluated on single click
        doubleClickActionBlock  <Block>    action evaluated on double click
        clickPosition           <Point>    internal use
        highlightMode           <Symbol>   how to draw the selection
        useIndex                <Boolean>  representation of the model selection
        hilightFgColor          <Color>           
        hilightBgColor          <Color>    how highlighted items are drawn
        hilightLevel            <Integer>  level to draw selections (i.e. for 3D effect)
        hilightFrameColor       <Color>    rectangle around highlighted items
        hilightStyle            <Boolean>  actions on widget are enabled/disabled

    [author:]
        Claus Atzkern

    [see also:]

        ListModelView
        HierarchicalListView
"


!

examples
"
                                                                        [exBegin]
    |top list view|

    list := List new.

    1 to:100 do:[:i| list add:('element: ', i printString) ].
    top  := StandardSystemView new; extent:300@300.
    view := ScrollableView for:SelectionInListModelView miniScroller:true
                        origin:0.0@0.0 corner:1.0@1.0 in:top.
    view list:list.
    top  open.
                                                                        [exEnd]


                                                                        [exBegin]
    |top list view item|

    list := HierarchicalList new.
    item := HierarchicalItem::Example labeled:'Test'.
    item expand.
    list showRoot:false.
    list root:item.

    top  := StandardSystemView new; extent:300@300.
    view := ScrollableView for:SelectionInListModelView miniScroller:true
                        origin:0.0@0.0 corner:1.0@1.0 in:top.

    view list:list.
    view doubleClickAction:[:i| (list at:i) toggleExpand ].
    top  open.
                                                                        [exEnd]

"

! !

!SelectionInListModelView class methodsFor:'initialization'!

initialize
    "setup the signals
    "
    super initialize.

    DefaultHilightLevel isNil ifTrue:[
        self updateStyleCache
    ].


! !

!SelectionInListModelView class methodsFor:'defaults'!

updateStyleCache
    "extract values from the styleSheet and cache them in class variables"

    <resource: #style   (
                        #'selection.hilightForegroundColor' #'selection.hilightBackgroundColor'
                        #'selection.hilightFrameColor'      #'selection.hilightLevel'
                        #'selection.foregroundColor'        #'selection.backgroundColor'
                        #'selection.shadowColor'            #'selection.lightColor'
                        #'selection.font'                   #'selection.hilightStyle'
                        #'text.foregroundColor'
                        )>

    DefaultHilightForegroundColor  := StyleSheet colorAt:'selection.hilightForegroundColor'.
    DefaultHilightBackgroundColor  := StyleSheet colorAt:'selection.hilightBackgroundColor'.
    DefaultHilightFrameColor       := StyleSheet colorAt:'selection.hilightFrameColor'.
    DefaultHilightLevel            := StyleSheet at:'selection.hilightLevel' default:0.
    DefaultHilightStyle            := StyleSheet at:'selection.hilightStyle' default:(StyleSheet name).
    DefaultForegroundColor         := StyleSheet colorAt:'selection.foregroundColor'.
    DefaultBackgroundColor         := StyleSheet colorAt:'selection.backgroundColor'.
    DefaultShadowColor             := StyleSheet colorAt:'selection.shadowColor'.
    DefaultLightColor              := StyleSheet colorAt:'selection.lightColor'.

    DefaultForegroundColor isNil ifTrue:[
        DefaultForegroundColor := StyleSheet colorAt:'text.foregroundColor' default:Black
    ].
    "
     self updateStyleCache
    "


! !

!SelectionInListModelView methodsFor:'accessing'!

list:aList
    "get the status of <showRoot> from the list
    "
    self deselectWithoutRedraw.
  ^ super list:aList
! !

!SelectionInListModelView methodsFor:'accessing behavior'!

highlightMode
    "get the mode how to draw a selected line:
        #line           draw whole line selected
        #label          draw label selected
    "
    ^ highlightMode


!

highlightMode:aMode
    "set the mode how to draw a selected line:
        #line           draw whole line selected
        #label          draw label selected
    "
    aMode notNil ifTrue:[
        highlightMode := aMode.
        self hasSelection ifTrue:[self invalidate]
    ]
!

multipleSelectOk
    "allow/disallow multiple selections; the default is false
    "
    ^ multipleSelectOk


!

multipleSelectOk:aState
    "allow/disallow multiple selections; the default is false
    "
    multipleSelectOk := aState ? false.
!

useIndex
    "set/clear the useIndex flag.
     the selection writen to the model are the indices into the list
     or the elements selected.
    "
    ^ useIndex


!

useIndex:aBoolean
    "set/clear the useIndex flag.
     the selection writen to the model are the indices into the list
     or the elements selected.
    "
    useIndex := aBoolean ? true


! !

!SelectionInListModelView methodsFor:'actions'!

action
    "get the action block to be performed on select
    "
    ^ actionBlock
!

action:aOneArgAction
    "set the action block to be performed on select
    "
    actionBlock := aOneArgAction


!

doubleClickAction
    "get the action block to be performed on doubleclick
    "
    ^ doubleClickActionBlock

!

doubleClickAction:aOneArgAction
    "set the action block to be performed on doubleclick
    "
    doubleClickActionBlock := aOneArgAction

! !

!SelectionInListModelView methodsFor:'change & update'!

argForChangeMessage
    "return the argument for a selectionChange;
     depending on the setting of useIndex, this is either the numeric
     index of the selection or the value (i.e. the string)
    "
    selection isNil ifTrue:[
        multipleSelectOk ifTrue:[^ #()].
        ^ useIndex ifTrue:[0] ifFalse:[nil]
    ].
    useIndex ifTrue:[^ selection copy].

    multipleSelectOk ifFalse:[
        ^ list at:selection ifAbsent:nil
    ].
    ^ selection collect:[:nr| list at:nr ifAbsent:nil ]
!

getSelectionFromModel
    "get selection from model; returns a selection or nil
    "
    |value nsel|

    ((value := model value) isNil or:[value == 0]) ifTrue:[
        ^ nil
    ].

    multipleSelectOk ifFalse:[
        "/ SINGLE SELECTION

        useIndex ifFalse:[
            value := list identityIndexOf:value.
            value == 0 ifTrue:[^ nil].
        ].
        ^ value
    ].

    "/ MULTI SELECT

    value isCollection ifFalse:[
        useIndex ifFalse:[
            value := list identityIndexOf:value.
            value == 0 ifTrue:[^ nil].
        ].
        ^ Array with:value
    ].

    value isEmpty ifTrue:[^ nil].

    useIndex ifTrue:[
        ^ value copy
    ].

    nsel := OrderedCollection new.
    value do:[:e||i|(i := list identityIndexOf:e) ~~ 0 ifTrue:[nsel add:i]].
  ^ nsel notEmpty ifTrue:[nsel] ifFalse:[nil]
!

listSizeChanged:aLnNr nLines:aDeltaLines
    "update selection
    "
    |newSel noChg size changed|

    super listSizeChanged:aLnNr nLines:aDeltaLines.

    selection isNil ifTrue:[^ self].

    list size == 0 ifTrue:[
      ^ self deselectWithoutRedraw
    ].
    multipleSelectOk ifFalse:[
        selection >= aLnNr ifTrue:[
            selection := selection + aDeltaLines.

            (aDeltaLines < 0 and:[selection < aLnNr]) ifTrue:[
                self deselectWithoutRedraw
            ] ifFalse:[
                (model notNil and:[useIndex]) ifTrue:[
                    model setValue:selection
                ]
            ]
        ].
        ^ self
    ].

    size    := selection size.
    changed := false.

    aDeltaLines < 0  ifFalse:[
        1 to:size do:[:anIndex|
            newSel := selection at:anIndex.

            newSel >= aLnNr ifTrue:[
                changed := true.
                selection at:anIndex put:(newSel + aDeltaLines)
            ]
        ].
        (changed and:[useIndex and:[model notNil]]) ifTrue:[
            model setValue:(selection copy)
        ].
        ^ self
    ].
    noChg := 0.

    1 to:size do:[:anIndex|
        newSel := selection at:anIndex.

        newSel >= aLnNr ifTrue:[
            newSel  := newSel + aDeltaLines.
            changed := true.

            newSel < aLnNr ifTrue:[
                noChg  := noChg + 1.
                newSel := 0.
            ].
            selection at:anIndex put:newSel
        ]
    ].

    noChg ~~ 0 ifTrue:[
        noChg == size ifTrue:[
            self deselectWithoutRedraw.
        ] ifFalse:[
            selection := selection select:[:i| i ~~ 0].
            self selectionChanged
        ]
    ] ifFalse:[
        (changed and:[useIndex and:[model notNil]]) ifTrue:[
            model setValue:(selection copy)
        ]
    ]

!

selectionChanged
    "selection has changed. Call actionblock and/or send changeMessage if defined
    "
    |arg|

    (model isNil and:[actionBlock isNil]) ifTrue:[
        ^ self
    ].

    arg := self argForChangeMessage.

    model notNil ifTrue:[
        model removeDependent:self.
        self sendChangeMessage:#value: with:arg.
        model addDependent:self.
    ].

    actionBlock notNil ifTrue:[
        (actionBlock numArgs) == 1 ifTrue:[
            actionBlock value:arg
        ] ifFalse:[
            actionBlock value
        ]
    ].

!

update:something with:aParameter from:changedObject
    "check wether my model changed
    "
    |newSelection|

    changedObject == model ifTrue:[
        self syncPushEvent:#updateFromModel
    ] ifFalse:[
        super update:something with:aParameter from:changedObject
    ].

!

updateFromModel
    "update selection from the model
    "
    model notNil ifTrue:[
        self syncFlushEvent:#updateFromModel.
        self setSelection:(self getSelectionFromModel).
    ]
! !

!SelectionInListModelView methodsFor:'drag & drop'!

canDrag
    "returns true if dragging is enabled
    "
    ^ dropSource notNil

!

dropSource
    "returns the dropSource or nil
    "
    ^ dropSource

!

dropSource:aDropSourceOrNil
    "set the dropSource or nil
    "
    dropSource := aDropSourceOrNil.

!

dropTarget
    "returns the dropTarget or nil
    "
    ^ dropTarget


!

dropTarget:aDropTragetOrNil
    "set the dropTarget or nil
    "
    dropTarget := aDropTragetOrNil.

!

startDragAt:aPoint
    "start drag at a point
    "
    dropSource notNil ifTrue:[
        dropSource startDragSelector notNil ifTrue:[
            dropSource startDragIn:self at:aPoint
        ] ifFalse:[
            DragAndDropManager new startDragFrom:self
                                      dropSource:dropSource
                                          offset:#center
        ]
    ]

! !

!SelectionInListModelView methodsFor:'drawing'!

drawFrom:start to:stop x:x y:y w:w
    "draw the lines between start to stop without clearing the background
    "
    highlightMode notNil ifTrue:[
        self selectionDo:[:lnNr|
            (lnNr between:start and:stop) ifTrue:[
                self drawSelectionFrameAt:lnNr x:x w:w
            ]
        ]
    ].
    self drawElementsFrom:start to:stop x:x y:y w:w.


!

drawLabelAt:anIndex x:x y:y h:h
    "draw the label at position x/y without clearing the background
    "
    |item|

    item := list at:anIndex ifAbsent:nil.

    item notNil ifTrue:[
        (highlightMode notNil and:[self isInSelection:anIndex]) ifTrue:[
            self paint:hilightFgColor on:hilightBgColor
        ] ifFalse:[
            self paint:fgColor on:bgColor.
        ].
        self displayElement:item x:x y:y h:h.
    ]


!

drawSelectionFrameAt:lnNr x:x w:w
    "draw the background and foreground of the selection frame
     at a lineNr.
    "
    |item
     x0 "{ Class:SmallInteger }"
     x1 "{ Class: SmallInteger }"
     y0 "{ Class:SmallInteger }"
     y1 "{ Class:SmallInteger }"
     hL "{ Class:SmallInteger }"
     wL "{ Class:SmallInteger }"
    |
    (item := list at:lnNr ifAbsent:nil) isNil ifTrue:[
        ^ self
    ].

    "/ CLEAR THE BACKGROUND

    y0 := self yVisibleOfLine:lnNr.
    y1 := self yVisibleOfLine:(lnNr + 1).
    hL := y1 - y0.

    highlightMode == #line ifTrue:[
        x0 := x.
        x1 := x0 + w.
    ] ifFalse:[
        highlightMode == #label ifFalse:[^ self].
        x0 := (self xVisibleOfItem:item) - (textStartLeft // 2).
        x1 := x0 + (item widthOn:self) + textStartLeft.
    ].
    x0 := x0 max:(x - 1).
    x1 := x1 min:width.
    wL := x1 - x0.

    wL > 0 ifFalse:[^ self].

    self paint:hilightBgColor.
    self fillRectangleX:x0 y:y0 width:wL height:hL.

    "/ DRAW THE FRAME

    hilightFrameColor notNil ifTrue:[
        hilightLevel == 0 ifTrue:[
            self paint:hilightFrameColor.

            highlightMode == #line ifTrue:[
                self displayLineFromX:x0 y:y0 toX:x1 y:y0.
                y1 := y0 + hL - 1.
                self displayLineFromX:x0 y:y1 toX:x1 y:y1.
            ] ifFalse:[
                self displayRectangleX:x0 y:y0 width:wL height:hL
            ].
            ^ self.
        ]
    ] ifFalse:[
        hilightStyle == #motif ifTrue:[
            self paint:bgColor.
            y1 := y0 + 1.
            highlightMode == #line ifTrue:[
                self displayLineFromX:x0 y:y1 toX:x1 y:y1.
                y1 := y0 + hL - 2.
                self displayLineFromX:x0 y:y1 toX:x1 y:y1.
            ] ifFalse:[
                self displayRectangleX:x0 + 1 y:y1 width:wL - 2 height:hL - 2
            ]
        ]
    ].

    hilightLevel ~~ 0 ifTrue:[
        "/ draw edge
        highlightMode == #line ifTrue:[
            x0 := margin.
            wL := width - x0 - x0.
        ].
        self drawEdgesForX:x0 y:y0 width:wL height:hL level:hilightLevel.
    ]

!

redrawSelectionAt:aLineNr
    "called to redraw a line caused by a change of the selection
    "
    highlightMode notNil ifTrue:[
        self redrawLineAt:aLineNr
    ].

! !

!SelectionInListModelView methodsFor:'event handling'!

buttonMotion:buttonMask x:x y:y
    "mouse-move while button was pressed - handle selection changes
    "
    |idx  p cY oY lnNr h|

    (enabled and:[self hasSelection]) ifFalse:[^ self].

    dragAccessPoint notNil ifTrue:[
        p := x @ y.

        (dragAccessPoint dist:p) > 5.0 ifTrue:[
            dragAccessPoint := nil.
            self startDragAt:p
        ].
        ^ self
    ].

    (multipleSelectOk and:[self sensor leftButtonPressed]) ifFalse:[
        ^ self
    ].

    clickLine isNil ifTrue:[
        ^ self
    ].
    cY := self yVisibleOfLine:clickLine.
    oY := cY.

    y < cY ifTrue:[
        (lnNr := clickLine - 1) == 0 ifTrue:[^ self].
        cY := self yVisibleOfLine:lnNr.
        h  := oY - cY.
    ] ifFalse:[
        (    (lnNr := clickLine + 1) > list size
         or:[(cY   := self yVisibleOfLine:lnNr) > y]
        ) ifTrue:[
            ^ self
        ].
        h  := cY - oY.
    ].
    (self isInSelection:lnNr) ifTrue:[
        self removeFromSelection:lnNr
    ] ifFalse:[
        self addToSelection:lnNr
    ].
    clickLine := lnNr.

    (cY between:margin and:(height - h)) ifFalse:[
        self scrollToLine:lnNr.
    ].
    self selectionChanged.
!

buttonMultiPress:button x:x y:y
    "button was pressed multiple - handle a doubleClick action
    "
    clickLine := nil.
    dragAccessPoint := nil.

    enabled ifFalse:[^ self].

    ((button == 1) or:[button == #select]) ifFalse:[
        ^ super buttonMultiPress:button x:x y:y
    ].
    self doubleClicked
!

buttonPress:button x:x y:y
    "a button was pressed - handle selection here
    "
    |sensor nsel start step|

    clickLine       := nil.
    dragAccessPoint := nil.

    enabled ifFalse:[^ self].

    (button == 1 or:[button == #select]) ifFalse:[
        ^ super buttonPress:button x:x y:y
    ].

    (clickLine := self yVisibleToLineNr:y) isNil ifTrue:[
        ^ self
    ].

    (multipleSelectOk and:[(sensor := self sensor) notNil]) ifTrue:[
        sensor ctrlDown ifTrue:[
            self toggleSelection:clickLine.
          ^ self selectionChanged
        ].

        (sensor shiftDown and:[(start := self firstInSelection) notNil]) ifTrue:[
            step := clickLine < start ifTrue:[-1] ifFalse:[1].
            nsel := OrderedCollection new.

            start to:clickLine by:step do:[:i|nsel add:i].

            (self selectWithoutScroll:nsel redraw:true) ifTrue:[
                self selectionChanged
            ].
            ^ self
        ]
    ].

    (self canDrag and:[self isInSelection:clickLine]) ifTrue:[
        dragAccessPoint := x @ y
    ] ifFalse:[
        (self selectWithoutScroll:clickLine redraw:true) ifTrue:[
            self selectionChanged
        ]
    ]

!

buttonRelease:button x:x y:y
    "a button was released
    "
    enabled ifTrue:[ 
        (dragAccessPoint notNil and:[clickLine notNil]) ifTrue:[
            (self selectWithoutScroll:clickLine redraw:true) ifTrue:[
                self selectionChanged
            ]
        ]
    ].
    clickLine       := nil.
    dragAccessPoint := nil.


!

characterPress:aKey x:x y:y
    " a character is pressed - lookup and change selection
    "
    |lnNr size idx sensor stp to1 fr2|

    (enabled and:[(size := self size) > 1]) ifFalse:[
        ^ self
    ].
    lnNr := self firstInSelection ? 0.

    ((sensor := self sensor) notNil and:[sensor shiftDown]) ifTrue:[
        stp := -1.              "/ search backward
        to1 :=  1.
        fr2 := size.
    ] ifFalse:[
        stp := 1.               "/ search forward
        to1 := size.
        fr2 := 1.
    ].

    idx := self findLineFrom:lnNr+stp to:to1 by:stp startingWithCharacter:aKey.

    idx == 0 ifTrue:[
        idx := self findLineFrom:fr2 to:lnNr-stp by:stp startingWithCharacter:aKey
    ].
    idx ~~ 0 ifTrue:[^ self selection:idx]
!

doubleClicked
    "handle a double click
    "
    (doubleClickActionBlock notNil and:[self numberOfSelections == 1]) ifTrue:[
        (doubleClickActionBlock numArgs == 1) ifTrue:[
            doubleClickActionBlock value:(self selectedIndex)
        ] ifFalse:[
            doubleClickActionBlock value
        ]
    ]

!

findLineFrom:aStart to:aStop by:aStep startingWithCharacter:aCharacter
    "find a line starting with a character
    "
    |item char lbl cmp
     size     "{ Class:SmallInteger }"
     start    "{ Class:SmallInteger }"
     stop     "{ Class:SmallInteger }"
    |
    (size := list size) ~~ 0 ifTrue:[
        aStep > 0 ifTrue:[
            aStart > aStop ifTrue:[^ 0].
        ] ifFalse:[
            (aStep == 0 or:[aStop > aStart]) ifTrue:[^ 0]
        ].

        start := aStart < 0 ifTrue:[1] ifFalse:[aStart min:size].
        stop  := aStop  < 0 ifTrue:[1] ifFalse:[aStop  min:size].
        char  := aCharacter asUppercase.

        start to:stop by:aStep do:[:anIndex|
            item := list at:anIndex ifAbsent:nil.
            item isNil ifTrue:[^ 0].    "/ list changed

            lbl := item perform:#string ifNotUnderstood:nil.

            lbl notNil ifTrue:[
                cmp := lbl string at:1 ifAbsent:nil.

                cmp notNil ifTrue:[
                    (char == cmp or:[char == cmp asUppercase]) ifTrue:[
                        ^ anIndex
                    ]
                ]
            ]
        ]
    ].
    ^ 0


!

keyPress:aKey x:x y:y
    "a key was pressed - handle page-keys here
    "
    <resource: #keyboard( #Return #CursorUp #CursorDown )>

    |sensor n size lineNr|

    enabled ifFalse:[
        ^ super keyPress:aKey x:x y:y
    ].
    aKey == #Return ifTrue:[
        self numberOfSelections == 1 ifTrue:[self doubleClicked].
      ^ self
    ].

    aKey isCharacter ifTrue:[
        ^ self characterPress:aKey x:x y:y
    ].

    (aKey == #CursorUp or:[aKey == #CursorDown]) ifFalse:[
        ^ super keyPress:aKey x:x y:y
    ].

    (size := self size) == 0 ifTrue:[
        ^ self
    ].

    lineNr := self selectedIndex.
    sensor := self sensor.

    sensor notNil ifTrue:[
        n := (1 + (sensor compressKeyPressEventsWithKey:aKey)) \\ size.
        n == 0 ifTrue:[^ self].
    ] ifFalse:[
        n := 1
    ].

    aKey == #CursorUp ifTrue:[
        lineNr == 0 ifTrue:[lineNr := size + 1].
        (n := lineNr - n) <= 0 ifTrue:[n := size + n]
    ] ifFalse:[
        (n := lineNr + n) > size ifTrue:[n := n - size]
    ].
    self selection:n
! !

!SelectionInListModelView methodsFor:'focus handling'!

wantsFocusWithPointerEnter
    "return true, if I want the focus when
     the mouse pointer enters"

    |pref|

    pref := UserPreferences current focusFollowsMouse.
    (pref ~~ false
    and:[(styleSheet at:#'selection.requestFocusOnPointerEnter' default:true)
    ]) ifTrue:[
        ^ true
    ].
    ^ false


! !

!SelectionInListModelView methodsFor:'initialize / release'!

fetchResources
    "fetch device colors and ..., to avoid reallocation at redraw time;
     *** called after a create or snapin to fetch all device resources
    "

    super fetchResources.

    hilightFgColor    := self colorOnDevice:hilightFgColor.
    hilightBgColor    := self colorOnDevice:hilightBgColor.
    hilightFrameColor := self colorOnDevice:hilightFrameColor.
!

initStyle
    "setup viewStyle specifics
    "
    |h|

    super initStyle.

    hilightFrameColor := nil.
    hilightLevel      := 0.
    hilightStyle      := DefaultHilightStyle.
    highlightMode     := #label.
    textStartLeft     := 4.

    device hasGrayscales ifTrue:[
        "
         must get rid of these hard codings
        "
        (hilightStyle == #next) ifTrue:[
            hilightFgColor := fgColor.
            hilightBgColor := White.
            hilightFrameColor := fgColor
        ] ifFalse:[
            (hilightStyle == #motif) ifTrue:[
                fgColor := White.
                bgColor := Grey.
                viewBackground := bgColor.
                hilightFgColor := bgColor.
                hilightBgColor := fgColor.
            ] ifFalse:[
                (hilightStyle == #openwin) ifTrue:[
                    hilightFgColor := fgColor.
                    hilightBgColor := Color grey.
                ]
            ]
        ]
    ].

    hilightFgColor isNil ifTrue:[
        hilightFgColor := bgColor.
    ].
    hilightBgColor isNil ifTrue:[
        hilightBgColor := fgColor.
    ].
    DefaultForegroundColor notNil ifTrue:[
        fgColor := DefaultForegroundColor
    ].
    DefaultBackgroundColor notNil ifTrue:[
        bgColor := viewBackground := DefaultBackgroundColor
    ].

    DefaultHilightForegroundColor notNil ifTrue:[
        hilightFgColor := DefaultHilightForegroundColor
    ].
    DefaultHilightBackgroundColor notNil ifTrue:[
        hilightBgColor := DefaultHilightBackgroundColor
    ].
    DefaultHilightFrameColor notNil ifTrue:[
        hilightFrameColor := DefaultHilightFrameColor
    ].
    DefaultHilightLevel notNil ifTrue:[
        hilightLevel := DefaultHilightLevel
    ].
    lineSpacing := 2 * (hilightLevel abs).

    (hilightStyle == #motif) ifTrue:[
        lineSpacing := lineSpacing max:6.
    ] ifFalse:[
        lineSpacing := lineSpacing max:4.
    ].
    hilightFgColor isNil ifTrue:[
        hilightFgColor := bgColor.
        hilightBgColor := fgColor
    ].
!

initialize
    "setup default attributes/behavior
    "
    super initialize.

    multipleSelectOk := false.
    useIndex         := true.

!

realize
    "get selection from model; scroll to selection
    "
    self updateFromModel.
    super realize.
    self makeSelectionVisible.

! !

!SelectionInListModelView methodsFor:'protocol'!

drawElementsFrom:start to:stop x:x y:y w:w
    "draw the items between start to stop without clearing the background
    "
    |y0 "{ Class:SmallInteger }"
     y1 "{ Class:SmallInteger }"
     x0 "{ Class:SmallInteger }"
    |
    x0 := textStartLeft - viewOrigin x.
    y1 := y.

    start to:stop do:[:i|
        y0 := y1.
        y1 := self yVisibleOfLine:(i + 1).
        self drawLabelAt:i x:x0 y:y0 h:(y1 - y0)
    ].


!

lostSynchronisation
    "called when the changes derived from the model are faster than the handling
    "
    self deselectWithoutRedraw.
    super lostSynchronisation.
!

xVisibleOfItem:anItem
    "returns the visible x of the labeled text
    "
    ^ textStartLeft - viewOrigin x


! !

!SelectionInListModelView methodsFor:'selection'!

deselect
    "clear selection
    "
    self selection:nil

!

firstInSelection
    "returns line number of first element selected or nil
    "
    multipleSelectOk ifFalse:[
        ^ selection
    ].
    ^ selection isNil ifTrue:[nil] ifFalse:[selection at:1]
!

hasSelection
    "returns true if a selection exists
    "
    ^ selection notNil

!

isInSelection:aNumber
    "return true, if line, aNumber is in the selection
    "
    multipleSelectOk ifFalse:[
        ^ aNumber == selection
    ].
    ^ selection isNil ifTrue:[false]
                     ifFalse:[selection includesIdentical:aNumber]
!

numberOfSelections
    "return the number of selected items
    "
    selection notNil ifTrue:[
        ^ multipleSelectOk ifFalse:[1] ifTrue:[selection size]
    ].
    ^ 0

!

selectElement:anElement
    "select the element. Scroll to make the new selection visible.
     Model and/or actionBlock notification IS done.
    "
    self selection:(list identityIndexOf:anElement)
!

selectedElement
    "return the single selected item or nil
    "
    |index|

    (index := self selectedIndex) ~~ 0 ifTrue:[
        ^ list at:index ifAbsent:nil
    ].
    ^ nil
!

selectedIndex
    "returns the index of the selected line or 0. If more
     lines are selected, 0 is returned
    "
    self numberOfSelections == 1 ifTrue:[
        ^ self firstInSelection
    ].
    ^ 0
!

selection
    "return the selection index or collection of indices 
     in case of multiple selection enabled
    "
    ^ selection

!

selection:something
    "select something or deselect if the argument is nil;
     scroll to make the selected line visible.
     The model and/or actionBlock IS notified.
    "
    (self selectWithoutScroll:something redraw:true) ifTrue:[
        self makeSelectionVisible.
        self selectionChanged
    ].

!

selectionDo:aBlock
    "perform aBlock for each nr in the selection.
     For single selection, it is called once for the items nr.
     For multiple selections, it is called for each.
    "
    selection notNil ifTrue:[
        multipleSelectOk ifTrue:[
            selection do:aBlock
        ] ifFalse:[
            aBlock value:selection
        ].
    ].


!

setSelection:something
    "select something or deselect if the argument is nil;
     scroll to make the selected line visible.
     *** No model and/or actionBlock notification is done here.
    "
    (self selectWithoutScroll:something redraw:true) ifTrue:[
        self makeSelectionVisible
    ]

!

toggleSelection:aNumber
    "toggle selection-state of entry, aNumber.
     *** No model and/or actionBlock notification is done here.
    "
    (self isInSelection:aNumber) ifTrue:[
        self removeFromSelection:aNumber
    ] ifFalse:[
        self addToSelection:aNumber
    ]

! !

!SelectionInListModelView methodsFor:'selection private'!

addToSelection:aNumber
    "add a number to the selection. No scrolling is done.
     *** No model and/or actionBlock notification is done here.
    "
    |newSelect oldSelect|

    multipleSelectOk ifFalse:[
        oldSelect == selection ifTrue:[^ self].
        oldSelect := selection.
        selection := aNumber.
        oldSelect notNil ifTrue:[self redrawSelectionAt:oldSelect].
    ] ifTrue:[
        selection isNil ifTrue:[
            selection := OrderedCollection new
        ] ifFalse:[
            (selection includesIdentical:aNumber) ifTrue:[^ self].

            selection isArray ifTrue:[
                selection := selection asOrderedCollection
            ]
        ].
        selection add:aNumber
    ].
    self redrawSelectionAt:aNumber.
!

deselectWithoutRedraw
    "clear the selection without redraw and scrolling;
     the model and/or actionBlock is notified
    "
    (self selectWithoutScroll:nil redraw:false) ifTrue:[
        self selectionChanged
    ]
!

makeSelectionVisible
    "scroll to make the selection line visible
    "
    |lineNr|

    selection notNil ifTrue:[
        (lineNr := self firstInSelection) notNil ifTrue:[
            self scrollToLine:lineNr
        ]
    ]

!

removeFromSelection:aNumber
    "remove aNumber from the selection and redraw line;
     *** No model and/or actionBlock notification is done here.
    "
    selection notNil ifTrue:[
        multipleSelectOk ifTrue:[
            (selection includesIdentical:aNumber) ifTrue:[
                selection size == 1 ifTrue:[
                    selection := nil
                ] ifFalse:[
                    selection isArray ifTrue:[
                        selection := selection asOrderedCollection
                    ].
                    selection removeIdentical:aNumber
                ].
                self redrawSelectionAt:aNumber
            ]
        ] ifFalse:[
            aNumber == selection ifFalse:[
                selection := nil.
                self redrawSelectionAt:aNumber
            ]
        ]
    ]

!

selectWithoutScroll:something redraw:doRedraw
    "change selection to something; returns true if selection changed
    "
    |oldSelect|

    selection = something ifTrue:[^ false].

    oldSelect := selection.

    (something isNil or:[something == 0]) ifTrue:[
        selection := nil
    ] ifFalse:[
        something isNumber ifTrue:[
            selection := multipleSelectOk ifTrue:[Array with:something] ifFalse:[something]
        ] ifFalse:[
            something size == 0 ifTrue:[
                selection := nil
            ] ifFalse:[
                selection := multipleSelectOk ifTrue:[something]
                                             ifFalse:[something at:1]
            ]
        ]
    ].
    selection = oldSelect ifTrue:[^ false].

    (doRedraw and:[shown]) ifFalse:[
        ^ true
    ].

    multipleSelectOk ifFalse:[
        oldSelect notNil ifTrue:[self redrawSelectionAt:oldSelect].
        selection notNil ifTrue:[self redrawSelectionAt:selection].
    ] ifTrue:[
        (selection notNil and:[oldSelect notNil]) ifTrue:[
            selection do:[:i|(oldSelect includesIdentical:i) ifFalse:[self redrawSelectionAt:i]].
            oldSelect do:[:i|(selection includesIdentical:i) ifFalse:[self redrawSelectionAt:i]].
        ] ifFalse:[
            oldSelect isNil ifTrue:[oldSelect := selection].
            oldSelect do:[:i|self redrawSelectionAt:i]
        ]
    ].
    ^ true


! !

!SelectionInListModelView class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libwidg2/SelectionInListModelView.st,v 1.19 1999-09-23 08:18:38 cg Exp $'
! !
SelectionInListModelView initialize!
